
#ifndef CQRLIB_H_INCLUDED
#define CQRLIB_H_INCLUDED

#include <math.h>
#include "Vector_3d.h"

#include <climits>
#include <cmath>

template< typename DistanceType=double >
class QR
{
private:
    DistanceType w, x, y, z;

public:
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline QR( ) : // default constructor
        w( DistanceType( 0.0 ) ),
        x( DistanceType( 0.0 ) ),
        y( DistanceType( 0.0 ) ),
        z( DistanceType( 0.0 ) )
    {
    }

    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline QR( const QR& q ) // copy constructor
    {
        if ( this != &q )
        {
            w = q.w;
            x = q.x;
            y = q.y;
            z = q.z;
        }
    }

    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline QR(                  // constructor
        const DistanceType& wi,
        const DistanceType& xi, 
        const DistanceType& yi, 
        const DistanceType& zi 
        ) :
        w( wi ),
        x( xi ),
        y( yi ),
        z( zi )
    {
    }

    /* Set -- set the values of an existing quaternion = w +ix+jy+kz */
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline void Set ( const DistanceType& wi,
        const DistanceType& xi, 
        const DistanceType& yi, 
        const DistanceType& zi 
        ) 
    {
        w = wi;
        x = xi;
        y = yi;
        z = zi;
    }

    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline DistanceType GetW( void ) const
    {
        return( w );
    }


    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline DistanceType GetX( void ) const
    {
        return( x );
    }


    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline DistanceType GetY( void ) const
    {
        return( y );
    }


    inline DistanceType GetZ( void ) const
    {
        return( z );
    }

    /* Add -- add a quaternion (q1) to a quaternion (q2) */   
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline QR operator+ ( const QR& q ) const
    {
        QR temp;
        temp.w = w + q.w;
        temp.x = x + q.x;
        temp.y = y + q.y;
        temp.z = z + q.z;
        return( temp );
    }

    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline QR& operator+= ( const QR& q )
    {
        w += q.w;
        x += q.x;
        y += q.y;
        z += q.z;
    }

    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline QR& operator-= ( const QR& q )
    {
        w -= q.w;
        x -= q.x;
        y -= q.y;
        z -= q.z;
    }

    /* Subtract -- subtract a quaternion (q2) from a quaternion (q1)  */  
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline QR operator- ( const QR& q ) const
    {
        QR temp;
        temp.w = w - q.w;
        temp.x = x - q.x;
        temp.y = y - q.y;
        temp.z = z - q.z;
        return( temp );
    }

    /* Multiply -- multiply a quaternion (q1) by quaternion (q2)  */    
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline QR operator* ( const QR& q ) const // multiply two quaternions
    {
        QR temp;
        temp.w = -z*q.z - y*q.y - x*q.x + w*q.w;
        temp.x =  y*q.z - z*q.y + w*q.x + x*q.w;
        temp.y = -x*q.z + w*q.y + z*q.x + y*q.w;
        temp.z =  w*q.z + x*q.y - y*q.x + z*q.w;
        return( temp );
    }

    /* Divide -- Divide a quaternion (q1) by quaternion (q2)  */
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    QR operator/ ( const QR& q2 ) const
    {
        const DistanceType norm2sq = q2.w*q2.w + q2.x*q2.x + q2.y*q2.y + q2.z*q2.z;

        if ( norm2sq == 0.0 )
        {
            return( QR( DBL_MAX, DBL_MAX, DBL_MAX, DBL_MAX ) );
        }

        QR q;
        q.w =  z*q2.z + y*q2.y + x*q2.x + w*q2.w;
        q.x = -y*q2.z + z*q2.y - w*q2.x + x*q2.w;
        q.y =  x*q2.z - w*q2.y - z*q2.x + y*q2.w;
        q.z = -w*q2.z - x*q2.y + y*q2.x + z*q2.w;

        return( q / norm2sq );
    }

    /* ScalarMultiply -- multiply a quaternion (q) by scalar (s)  */
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline QR operator* ( const DistanceType& d ) const // multiply by a constant
    {
        QR temp;
        temp.w = w*d;
        temp.x = x*d;
        temp.y = y*d;
        temp.z = z*d;
        return( temp );
    }

    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline QR operator/ ( const DistanceType& d ) const // divide by a constant
    {
        QR temp;
        temp.w = w/d;
        temp.x = x/d;
        temp.y = y/d;
        temp.z = z/d;
        return( temp );
    }

    /* Conjugate -- Form the conjugate of a quaternion qconj */
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline QR Conjugate ( void ) const
    {
        QR conjugate;
        conjugate.w =  w;
        conjugate.x = -x;
        conjugate.y = -y;
        conjugate.z = -z;
        return( conjugate );
    }

    /* Normsq -- Form the normsquared of a quaternion */
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline DistanceType Normsq ( void ) const
    {
        return( w*w + x*x + y*y + z*z );
    }

    /* Inverse -- Form the inverse of a quaternion */
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline QR Inverse ( void) const
    {
        QR inversequaternion = (*this).Conjugate();
        const DistanceType normsq = (*this).Normsq( );

        if ( normsq > DistanceType( 0.0 ) )
        {
            inversequaternion = inversequaternion * DistanceType( 1.0 ) / normsq;
        }
        else
        {
            inversequaternion = QR( DBL_MAX, DBL_MAX, DBL_MAX, DBL_MAX );
        }
        return( inversequaternion );
    }

    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline QR& operator= ( const QR& q ) 
    {
        if ( this != &q )
        {
            w = q.w;
            x = q.x;
            y = q.y;
            z = q.z;
        }
        return( *this );
    }

    /* Equal */    
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline bool operator== ( const QR& q ) const
    {
        return( w==q.w && x==q.x && y==q.y && z==q.z );
    }

    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline bool operator!= ( const QR& q ) const
    {
        return( (w!=q.w) || (x!=q.x) || (y!=q.y) | (z!=q.z) );
    }

    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline Vector_3 operator* ( const Vector_3& v )
    {
        return( QRRotateByQuaternion( v ) );
    }

    /* RotateByQuaternion -- Rotate a vector by a Quaternion, w = qvq* */
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    static inline Vector_3 QRRotateByQuaternion( const Vector_3& v )
    {
        QR vquat( 0.0, v[0], v[1], v[2] );
        const QR wquat = (*this)*vquat;
        const QR qconj = *this.Conjugate( );
        vquat = wquat * qconj;
        return( Vector_3( vquat.x, vquat.y, vquat.z ) );
    }

    /* Axis2Quaternion -- Form the quaternion for a rotation around axis v  by angle theta */
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    static inline QR QRAxis2Quaternion ( const DistanceType& angle, const Vector_3& v )
    {
        return( QRAxis2Quaternion( v, angle ) );
    }

    /* Axis2Quaternion -- Form the quaternion for a rotation around axis v  by angle theta */
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    static inline QR QRAxis2Quaternion ( const Vector_3& v, const DistanceType& angle  )
    {
        const DistanceType norm2sq = v.SquaredLength();

        if ( norm2sq == DistanceType(0.0) )
        {
            return( QR( DBL_MAX, DBL_MAX, DBL_MAX, DBL_MAX ) );
        }

        const DistanceType sinOverNorm = sin(angle/2.0)/sqrt(norm2sq);
        const QR q( cos(angle/2.0), sinOverNorm*v[0], sinOverNorm*v[1], sinOverNorm*v[2]);

        return( q );
    }

    /* Matrix2Quaterion -- Form the quaternion from a 3x3 rotation matrix R */
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    static inline QR QRMatrix2Quaternion ( const Matrix_3x3& m )
    {
        const DistanceType trace = m[0] + m[4] + m[8];
        QR rotquaternion;

        if ( trace > -0.75 )
        {
            rotquaternion.w = sqrt( (1.0+trace)) / 2.0;
            const DistanceType recip = 0.25 / rotquaternion.w;
            rotquaternion.z = (m[3]-m[1]) * recip;
            rotquaternion.y = (m[2]-m[6]) * recip;
            rotquaternion.x = (m[7]-m[5]) * recip;
            return( rotquaternion );
        }

        const DistanceType fourxsq = 1.0 + m[0] - m[4] - m[8];

        if ( fourxsq >= 0.25 )
        {
            rotquaternion.x = sqrt( fourxsq ) / 2.0;
            const DistanceType recip = 0.25 /rotquaternion.x;
            rotquaternion.y = m[3] + m[1]*recip;
            rotquaternion.z = m[2] + m[6]*recip;
            rotquaternion.w = m[7] + m[5]*recip;
            return( rotquaternion );
        }

        const DistanceType fourysq = 1.0 + m[4] - m[0] - m[8];

        if ( fourysq >= 0.25 )
        {
            rotquaternion.y = sqrt( fourysq ) / 2.0;
            const DistanceType recip = 0.25 / rotquaternion.y;
            rotquaternion.x = m[3] + m[1]*recip;
            rotquaternion.w = m[2] + m[6]*recip;
            rotquaternion.z = m[7] + m[5]*recip;
            return( rotquaternion );
        }

        const DistanceType fourzsq = 1.0 + m[8] - m[0] - m[4];

        if ( fourzsq >= 0.25 )
        {
            rotquaternion.z = sqrt( fourzsq ) / 2.0;
            const DistanceType recip = 0.25 / rotquaternion.z;
            rotquaternion.w = m[3] - m[1]*recip;
            rotquaternion.w = m[2] - m[6]*recip;
            rotquaternion.w = m[7] - m[6]*recip;
            return( rotquaternion );
        }

        return( QR( ) );
    }

    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    DistanceType operator[] ( const int k ) const
    {
        const int i = (k<0) ? 0 : ( (k>3)?3:k ) ;
        if ( i==0 ) return w;
        if ( i==1 ) return x;
        if ( i==2 ) return y;
        if ( i==3 ) return z;
        return( 0 ); // just to keep compilers happy
    }

    /* Quaternion2Matrix -- Form the 3x3 rotation matrix from a quaternion */    
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline Matrix_3x3 QRQuaternion2Matrix( void ) const
    {
        const DistanceType ww = w*w;
        const DistanceType xx = x*x;
        const DistanceType yy = y*y;
        const DistanceType zz = z*z;

        const DistanceType  twoxy = 2.0 * x*y;
        const DistanceType  twoyz = 2.0 * y*z;
        const DistanceType  twoxz = 2.0 * x*z;
        const DistanceType  twowx = 2.0 * w*x;
        const DistanceType  twowy = 2.0 * w*y;
        const DistanceType  twowz = 2.0 * w*z;

        const Matrix_3x3 m(
            ww+xx-yy-zz,
            twoxy - twowz,
            twoxz + twowy,

            twoxy + twowz,
            ww-xx+yy-zz,
            twoyz - twowx,

            twoxz - twowy,
            twoyz + twowx,
            ww-xx-yy+zz
            );

        return( m );
    } // end QRQuaternion2Matrix

    /* Get a unit quaternion from a general one */
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline QR UnitQ( void ) const
    {
        const DistanceType normsq = Normsq( );
        if ( normsq == 0.0 )
        {
            return( QR( DBL_MAX, DBL_MAX, DBL_MAX, DBL_MAX ) );
        }
        else
        {
            return( (*this) / sqrt( Normsq( ) ) );
        }
    } // end UnitQ

    /* Quaternion2Angles -- Convert a Quaternion into Euler Angles for Rz(Ry(Rx))) convention */  
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline bool QRQuaternion2Angles ( DistanceType& rotX, DistanceType& rotY, DistanceType& rotZ ) const
    {
        const DistanceType PI = 4.0 * atan( 1.0 );

        const DistanceType rmx0 = w*w + x*x - y*y - z*z;
        const DistanceType rmx1 = 2.0 * ( x*y - w*z );
        const DistanceType rmy0 = 2.0 * ( x*y + w*z );
        const DistanceType rmy1 = w*w - x*x + y*y - z*z;
        const DistanceType rmz0 = 2.0 * (x*z - w*y );
        const DistanceType rmz1 = 2.0 * (w*x + y*z );
        const DistanceType rmz2 = w*w - x*x - y*y + z*z;
        DistanceType srx, sry, srz, trX, trY, trZ;

        if ( rmz0 >= 1.0 )
        {
            sry = -0.5 * PI;
        }
        else if ( rmz0 <= -1.0 )
        {
            sry = 0.5 * PI;
        }
        else
        {
            sry = asin( -rmz0 );
        }

        if ( rmz0 > 0.9999995 )
        {
            srx = atan2( -rmx1, rmy1 );
            srz = 0.0;
        }
        else if ( rmz0 < -0.9999995 )
        {
            srx = atan2( rmx1, rmy1 );
            srz = 0.0;
        }
        else
        {
            srx = atan2( rmz1, rmz2 );
            srz = atan2( rmy0, rmx0 );
        }

        trX = PI + srx;
        if ( trX > 2.0 * PI ) trX -= 2.0 * PI;
        trY = PI + sry;
        if ( trY > 2.0 * PI ) trY -= 2.0 * PI;
        trZ = PI + srz;
        if ( trZ > 2.0 * PI ) trZ -= 2.0 * PI;

        const DistanceType nsum  =    fabs( cos(srx)-cos(rotX)) + fabs( sin(srx)-sin(rotX))
                                    + fabs( cos(sry)-cos(rotY)) + fabs( sin(sry)-sin(rotY))
                                    + fabs( cos(srz)-cos(rotZ)) + fabs( sin(srz)-sin(rotZ));

        const DistanceType tsum  =    fabs( cos(trX)-cos(rotX)) + fabs( sin(trX)-sin(rotX))
                                    + fabs( cos(trY)-cos(rotY)) + fabs( sin(trY)-sin(rotY))
                                    + fabs( cos(trZ)-cos(rotZ)) + fabs( sin(trZ)-sin(rotZ));

        if ( nsum < tsum )
        {
            rotX = srx; rotY = sry; rotZ = srz;
        }
        else
        {
            rotX = trX; rotY = trY; rotZ = trZ;
        }

        return( true );
    }  // end QRQuaternion2Angles

    /* Angles2Quaternion -- Convert Euler Angles for Rz(Ry(Rx))) convention into a quaternion */
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    static inline QR QRAngles2Quaternion ( const DistanceType& rotX, const DistanceType& rotY, const DistanceType& rotZ )
    {
        const DistanceType cx = cos( rotX / 2.0 );
        const DistanceType sx = sin( rotX / 2.0 );

        const DistanceType cy = cos( rotY / 2.0 );
        const DistanceType sy = sin( rotY / 2.0 );

        const DistanceType cz = cos( rotZ / 2.0 );
        const DistanceType sz = sin( rotZ / 2.0 );

        const QR q( cx*cy*cz + sx*sy*sz,
            sx*cy*cz - cx*sy*sz,
            cx*sy*cz + sx*cy*sz,
            cx*cy*sz - sx*sy*cz 
            ); 
        return( q );
    } // end QRAngles2Quaternion

    static inline QR Point2Quaternion( const Vector_3& v )
    {
        return( QR( 0.0, v[0], v[1], v[2] ) );
    }

}; // end class QR



#endif
