#include <complex>
#include <float.h>
#include <list>
#include <cmath>
#include <stdlib.h>
#ifdef USE_SYSTEM_HEADERS
#include <vector_3d.h>
#else
#include "vector_3d.h"
#endif

const double pi = 4.0 * ::atan( 1.0 );
const double degreesPerRadians = pi / 180.0;
const double RadiansPerDegree = 180.0 / pi;

void testCart( void );
void testVectors( void );
void testMatrices( void );
void testInteractions( void );
void testLinePlane( void );
void testEigen( void );
void testUnitFunctions( void );
void TestOnePair( const Vector_3& x1, const Vector_3& x2, const Vector_3& z1, const Vector_3& z2, int& localErrorCount );


const Vector_3 xaxis( 1.0, 0.0, 0.0 );
const Vector_3 yaxis( 0.0, 1.0, 0.0 );
const Vector_3 zaxis( 0.0, 0.0, 1.0 );

long g_errorCount;

int main (int argc, char ** argv)
{

   g_errorCount = 0;

    testCart( );
    testVectors( );
    testMatrices( );
    testInteractions( );
    testLinePlane( );
    testEigen( );
    testUnitFunctions( );

    if( g_errorCount == 0 )
    {
        fprintf(stdout,  "No errors were detected while testing Vector_3\n" );
    }

    return( g_errorCount );
}

/*
QUESTIONS:
should line and plane even be here?
should there be valid/invalid flags for the objects?
*/

/*=======================================================================*/
void testVectors( void )
{
    const Vector_3 v1(1.0);
    const Vector_3 v2( 2.0, 4.0, 8.0 );


    { /*test the access functions*/
        Vector_3 v( 1.0, 2.0, 3.0 );
        if( ::abs(v[-1] - v[0]) > 0.0 ||
            ::abs(v[2]  - v[3]) > 0.0   )
        {
            ++g_errorCount;
            fprintf( stdout, "testVectors: an access function failed\n" );
        }
    }
    { /*test Norm and SquaredLength*/
        if( ::abs(v2.Norm( ) - ::sqrt( (double)( 4+16+64 ) )) > 0.0 )
        {
            ++g_errorCount;
            fprintf( stdout, "testVectors: Norm failed\n" );
        }

        if( v2.SquaredLength( ) != (double)( 4+16+64 ) )
        {
            ++g_errorCount;
            fprintf( stdout, "StestVectors: quaredLength failed\n" );
        }

        if( ::abs(v2.Norm( )*v2.Norm( ) - v2.SquaredLength( )) > 0.0 )
        {
            ++g_errorCount;
            fprintf( stdout, "testVectors: Norm or SquaredLength failed \n" );
        }
    }

    { /*test addition and verify constructors*/
        const Vector_3 v3 = v1 + v2;
        if( v3.SquaredLength( ) != (double)(9+25+81) )
        {
            ++g_errorCount;
            fprintf( stdout, "testVectors: vector addition or Norm failed n" );
        }

        /*test subtraction*/
        const Vector_3 v4 = v1 - v2;
        const Vector_3 v5 = v2 - v1;
        if( v4.SquaredLength( ) + v5.SquaredLength( ) != (double)( 2*(1+9+49) ) )
        {
            ++g_errorCount;
            fprintf( stdout, "testVectors: vector subtraction failed \n" );
        }
    }

    { /*test the functions to get the principle axes*/
        Vector_3 vx;
        Vector_3 vy;
        Vector_3 vz;
        if(
            ( (Vector_3::GetXAxis( ) - xaxis).SquaredLength( ) != 0.0 ) ||
            ( (Vector_3::GetYAxis( ) - yaxis).SquaredLength( ) != 0.0 ) ||
            ( (Vector_3::GetZAxis( ) - zaxis).SquaredLength( ) != 0.0 )
            )
        {
            ++g_errorCount;
            fprintf( stdout, "testVectors: Get?Axis function failed \n" );
        }
    }

    { /*test vector operator +=  */
        Vector_3 v6( 1.0 );
        v6 += v2;
        if( v6.SquaredLength( ) != (double)( 9+25+81 ) )
        {
            ++g_errorCount;
            fprintf( stdout, "testVectors: vector operator+= failed for v2 \n" );
        }

        v6 += v1;
        if( v6.SquaredLength( ) != (double)( 16+36+100 ) )
        {
            ++g_errorCount;
            fprintf( stdout, "testVectors: vector operator+= failed for v1 \n" );
        }
    }

    { /*test vector operator -=  */
        Vector_3 v6( 1.0 );
        v6 -= v2;
        if( v6.SquaredLength( ) != (double)( 1+9+49 ) )
        {
            ++g_errorCount;
            fprintf( stdout, "testVectors: vector operator-= failed for v2 \n" );
        }

        v6 -= v1;
        if( v6.SquaredLength( ) != (double)( 4+16+64 ) )
        {
            ++g_errorCount;
            fprintf( stdout, "testVectors: vector operator-= failed for v1 \n" );
        }
    }

    { /*test vector/scalar functions*/
        /*test division*/
        const Vector_3 v = v2 / 2.0;
        if( v.SquaredLength( ) != (double)( 1+4+16 ) )
        {
            ++g_errorCount;
            fprintf( stdout, "testVectors: vector divided by scalar failed \n" );
        }

        /*test scalar times vector*/
        const Vector_3 v3 = 17.0 * v2;
        if( v3.SquaredLength( ) != 17.0 * 17.0 * (double)( 4+16+64 ) )
        {
            ++g_errorCount;
            fprintf( stdout, "testVectors: scalar times vector failed, result=%f\n", v3.SquaredLength( ) );
        }

        /*test vector times scalar*/
        const Vector_3 v4 = v2.operator*( 17.0 ); /*operator double causes ambiguity - should it be removed?*/
        if( v4.SquaredLength( ) != 17.0 * 17.0 * (double)( 4+16+64 ) )
        {
            ++g_errorCount;
            fprintf( stdout, "testVectors: vector times scalar failed, result=%f\n", v3.SquaredLength( ) );
        }
    }

    { /*Geometry*/
        /*test angle*/
        const double angle = Vector_3::Angle( xaxis, Vector_3::GetZeroVector( ), yaxis );
        if( ::fabs(angle-pi/2.0) >  10.0 * DBL_EPSILON ) /*allow for some error*/
        {
            ++g_errorCount;
            fprintf( stdout, "testVectors: Angle between x and y axes is incorrect, angle=%f(degrees)\n", angle/degreesPerRadians );
        }

        const double torsion = Vector_3::Torsion( xaxis, Vector_3::GetZeroVector( ), yaxis, zaxis );
        if( ::fabs( torsion+pi/2.0 ) > 10.0 * DBL_EPSILON ) /*allow for some error*/
        {
            ++g_errorCount;
            fprintf( stdout, "testVectors: Torsion angle is incorrect, angle=%f(degrees)\n", torsion/degreesPerRadians );
        }
    }

    { /*test operator dot*/
        const double d = v2.Dot( v2 );
        if( d != (double) ( 2*2+4*4+8*8 ) )
        {
            ++g_errorCount;
            fprintf( stdout, "testVectors: Dot product failed \n" );
        }

        const double dxx = Vector_3::GetXAxis( ).Dot( Vector_3::GetXAxis( ) );
        const double dyy = Vector_3::GetYAxis( ).Dot( Vector_3::GetYAxis( ) );
        const double dzz = Vector_3::GetZAxis( ).Dot( Vector_3::GetZAxis( ) );
        const double dxy = Vector_3::GetXAxis( ).Dot( Vector_3::GetYAxis( ) );
        const double dyz = Vector_3::GetYAxis( ).Dot( Vector_3::GetZAxis( ) );
        const double dxz = Vector_3::GetXAxis( ).Dot( Vector_3::GetZAxis( ) );
        const double dzx = Vector_3::GetZAxis( ).Dot( Vector_3::GetXAxis( ) );
        if(
            ::fabs( dxx - 1.0 ) > 1.0E-8 ||
            ::fabs( dyy - 1.0 ) > 1.0E-8 ||
            ::fabs( dzz - 1.0 ) > 1.0E-8 ||
            ::fabs( dxy ) > 1.0E-8 ||
            ::fabs( dxz ) > 1.0E-8 ||
            ::fabs( dyz ) > 1.0E-8 ||
            ::fabs( dzx ) > 1.0E-8  )
        {
            ++g_errorCount;
            fprintf( stdout, "testVectors: dot product failed \n" );;
        }

        const Vector_3 tiny( 1.0E-9 );
        int localErrorCount = 0;
        double multiplier = 0.001;
        double multiplierOfTiny = 0.1;
        for( int m=0; m<4; ++m )
        {
           //fprintf( stdout, "tiny multiplier(-) %g\n", multiplierOfTiny );
           multiplier = 0.001;
           for( int i=0; i<6; ++i )
           {
              //fprintf( stdout, "multiplier(-) %g\n", multiplier );
              TestOnePair( multiplier*xaxis, multiplier*yaxis, multiplier*(multiplierOfTiny*tiny-xaxis), multiplier*yaxis, localErrorCount );
              multiplier *= 10.0;
           }
           multiplierOfTiny *= 10.0;
        }

        multiplier = 0.001;
        for( int i=0; i<6; ++i )
        {
           //fprintf( stdout, "multiplier(+) %g\n", multiplier );
           TestOnePair( multiplier*xaxis, multiplier*yaxis, multiplier*(xaxis+tiny), multiplier*yaxis, localErrorCount );
           multiplier *= 10.0;
        }
        g_errorCount += localErrorCount;
    }

    { /*test cross product*/

        /*verify that cross products of the axes result in the third axis*/
        if(
            (xaxis.Cross( yaxis ).Dot( zaxis ) !=  1.0 ) ||
            (yaxis.Cross( xaxis ).Dot( zaxis ) != -1.0 ) ||
            (yaxis.Cross( zaxis ).Dot( xaxis ) !=  1.0 ) ||
            (zaxis.Cross( yaxis ).Dot( xaxis ) != -1.0 ) ||
            (zaxis.Cross( xaxis ).Dot( yaxis ) !=  1.0 ) ||
            (xaxis.Cross( zaxis ).Dot( yaxis ) != -1.0 )
            )
        {
            ++g_errorCount;
            fprintf( stdout, "testVectors: Cross product failed \n" );
        }

        /*verify that the cross product of a unit vector and one of size root 2
          at 45 degrees gives a unit vector (perpendicularity was verified above)*/
        const Vector_3 xy0( 1.0, 1.0, 0.0 );
        if(
            ::fabs((xaxis.Cross( xy0 )).SquaredLength( ) - 1.0 ) >
            10.0 * DBL_EPSILON  /*allow for some error*/
            )
        {
            ++g_errorCount;
            fprintf( stdout, "testVectors: Cross gave wrong magnitude, result=%f\n", (xaxis.Cross( xy0 )).SquaredLength( ) );
        }
    }

    { /*verify unary minus operator*/
        if( xaxis.Dot(-xaxis) != -1.0 )
        {
            ++g_errorCount;
            fprintf( stdout, "testVectors: unary minus operator failed \n" ) ;
        }
    }

    { /*test UnitV - Norm should be returned as 1.0*/
        if( ::fabs( v2.UnitV( ).SquaredLength( ) - 1.0 ) > 10.0 * DBL_EPSILON ) /*allow for some error*/
        {
            ++g_errorCount;
            fprintf( stdout, "testVectors: UnitV failed, result=%f\n", v2.UnitV( ).Norm( ) );
        }
    }

    { /*test DyadProduct*/
        const Vector_3 vDyad0(Vector_3(2.0,3.0,5.0) );
        const Vector_3 vDyad1(Vector_3(7.0,11.0,13.0) );
        const Vector_3 vDyad2 = vDyad0.DyadProduct(vDyad1);
        const double dTest = ::sqrt( vDyad0[0]*vDyad1[0]*vDyad0[0]*vDyad1[0] + vDyad0[1]*vDyad1[1]*vDyad0[1]*vDyad1[1] + vDyad0[2]*vDyad1[2]*vDyad0[2]*vDyad1[2] );
        if( ::fabs( vDyad2.Norm( ) - dTest ) > 10.0 * DBL_EPSILON ) /*allow for some error*/
        {
            ++g_errorCount;
            fprintf( stdout, "testVectors: DyadProduct failed \n" );
        }
    }
}

/*=======================================================================*/
void testMatrices( void )
{
    const Matrix_3x3 Unit(1.,0.,0., 0.,1.,0., 0.,0.,1.);


    { /*test the access functions*/
        Matrix_3x3 m( 1.0,2.0,3.0, 4.0,5.0,6.0, 7.0,8.0,9.0 );
        if( ::abs(m[-1] - m[0]) > 0.0 ||
            ::abs(m[8]  - m[9]) > 0.0   )
        {
            ++g_errorCount;
            fprintf( stdout, "testMatrices: an access function failed\n" );
        }
    }
    { /*test determinant*/
        if( ::fabs(Unit.Det( ) -1.0) > 1.0E-8 )
        {
            ++g_errorCount;
            fprintf( stdout, "testMatrices: determinant failed, result=%f\n", Unit.Det() );;
        }
    }
}

/*=======================================================================*/
void testInteractions( void )
{
    { /*test MatrixProduct*/
        const Vector_3 v1( 3.0, 5.0, 7.0 );
        const Vector_3 v2( 11.0, 13.0, 17.0 );
        const Matrix_3x3 m( v1.MatrixProduct(v2) );
        const Matrix_3x3 mTest( 33.0, 39.0, 51.0, 39.0, 65.0, 85.0, 51.0, 85.0, 119.0 );
        if( ::fabs( (mTest - m ).Det( ) ) > 10.0 * DBL_EPSILON ) /*allow for some error*/
        {
            ++g_errorCount;
            fprintf( stdout, "testInteractions: MatrixProduct failed\n" );;
        }
    }

    { /*test rotation matrix generation and vector-matrix multiply*/
        const Matrix_3x3 m2 = Vector_3::GetZAxis( ).Rotmat(45. * degreesPerRadians); /*do rotation about z axis*/
        const Vector_3 vo = Vector_3::GetXAxis( ).MV(m2); /*rotate the x axis*/
        const Vector_3 vTemp = vo - Vector_3( ::sqrt(2.0)/2.0, ::sqrt(2.0)/2.0,  0.0);
        if( vTemp.Norm( ) > 10.0 * DBL_EPSILON ) /*allow for some error*/
        {
            ++g_errorCount;
            fprintf( stdout, "testInteractions: rotation failed\n" );;
        }
    }

    { /*test Pair (which computes the rotation of one pair of vectors to another)*/
        const Vector_3 b(1.0, 1.0, 1.0);

        //const Matrix_3x3 mPair = Vector_3::Pair(Vector_3::GetXAxis( ),b, Vector_3::GetZAxis( ),Vector_3::GetYAxis( ));
        int localErrorCount = 0;
        TestOnePair(Vector_3::GetXAxis( ),b, Vector_3::GetZAxis( ),Vector_3::GetYAxis( ), localErrorCount);
        g_errorCount += localErrorCount;
        /*should result in xaxis being rotated to zaxis and 1,1,1 being rotated into the yz-plane*/
        //const Vector_3 vtest = mPair.MV(Vector_3::GetXAxis( ));
        //if( (vtest - Vector_3::GetZAxis( )).Norm( ) > 1.0E-8 )
        //{ /*Vector_3::GetXAxis( ) should be rotated onto the Vector_3::GetZAxis( )*/
        //    ++g_errorCount;
        //    fprintf( stdout, "testInteractions: Pair failed [1]\n");
        //}
        //if( ::fabs( ((Vector_3::GetZAxis( ).Cross(Vector_3::GetYAxis( ))).Dot(mPair.MV(b)))) > 1.0E-8 )
        //{ /*the rotated b-vector should be normal to the plane of Vector_3::GetZAxis( ),Vector_3::GetYAxis( )*/
        //    ++g_errorCount;
        //    fprintf( stdout, "testInteractions: Pair failed [2]\n");
        //}

        const Vector_3 vv1( xaxis );
        const Vector_3 vv2( 1.0, 1.0, 0.0 );
        const Vector_3 zz1( 1.0, 1.0, 1.0 );
        const Vector_3 zz2( 0.0, 1.0, 1.0 );

        //const Matrix_3x3 m = Vector_3::Pair( vv1, vv2, zz1, zz2 );
        localErrorCount = 0;
        TestOnePair( vv1, vv2, zz1, zz2, localErrorCount );
        g_errorCount += localErrorCount;
        //const double angle1 = Vector_3::Angle( m.MV( vv1 ), Vector_3::GetZeroVector( ), zz1 );

        //const double distance1 = Plane_3( zz1.Cross( zz2 ), Vector_3::GetZeroVector( )) .DistanceFromPlane( m.MV(vv1) );

        //if( distance1 > 2.0*DBL_EPSILON )
        //{
        //    ++g_errorCount;
        //    fprintf( stdout, "testInteractions: Pair distance1 not small enough\n" );
        //}

        //if( ::fabs(angle1)  > 10.0 * DBL_EPSILON ) /*allow for some error*/
        //{
        //    /*if this fails, vv1 was not rotated onto zz1*/
        //    ++g_errorCount;
        //    fprintf( stdout, "testInteractions: Pair failed for first rotation, angle1=%g\n", angle1/degreesPerRadians );
        //}

        //const double angle2 = Vector_3::Torsion( vv2.MV( m ), zz1, Vector_3::GetZeroVector( ), zz2 );
        //const double distance2 = zz1.Cross( zz2 ).UnitV( ).Dot(vv2.MV(m));
        //if( distance2 > 10.0*DBL_MIN )
        //{
        //    /*if this fails, vv2 was not rotated into the plane of zz1,0,zz2*/
        //    ++g_errorCount;
        //    fprintf( stdout, "testInteractions: Pair failed for second rotation, distance2=%g\n", distance2 );
        //}
        //else if( ::fabs(angle2)  > 10.0 * DBL_EPSILON ) /*allow for some error*/
        //{
        //    ++g_errorCount;
        //    fprintf( stdout, "testInteractions: Torsion failed for second rotation, angle2=%g degrees\n", angle2/degreesPerRadians );
        //}

    }

}

/*=======================================================================*/
void testLinePlane( void )
{

    { /*test best plane*/
        std::list<Vector_3> lv;
        lv.push_back( Vector_3::GetXAxis( ) );
        lv.push_back( Vector_3::GetYAxis( ) );
        lv.push_back( Vector_3::GetZAxis( ) );
        const Plane_3 vOut = BestPlane( lv );
        if( (vOut.NormalVector( ) - Vector_3(::sqrt(3.0), ::sqrt(3.0), ::sqrt(3.0) )/3.0 ).Norm( ) > 1.0E-8 )
        {
            ++g_errorCount;
            fprintf( stdout, "testLinePlane: BestPlane failed\n" );
        }
    }

    { /*test non-perfect best line thru origin*/
        std::list<Vector_3> lv;
        lv.push_back( Vector_3::GetXAxis( ) );
        lv.push_back( 2.0 * Vector_3::GetXAxis( ) );
        const Vector_3 delta(0.0,0.1,0.0);
        lv.push_back( Vector_3::GetXAxis( ) + delta );
        lv.push_back( Vector_3::GetXAxis( ) - delta );
        const Line_3 lOut = BestLine( lv );
        if( (lOut.LineAxis( ) - Vector_3::GetXAxis( )).Norm( ) > 1.0E-8 )
        {
            ++g_errorCount;
            fprintf( stdout, "testLinePlane: BestLine-1 Failed \n" /*<< lOut.LineAxis( )*/ );
        }
    }

    { /*test best line somewhere not thru the origin*/
        std::list<Vector_3> lv;
        lv.push_back( Vector_3::GetXAxis( ) );
        lv.push_back( Vector_3::GetXAxis( ) + Vector_3::GetYAxis( ) );
        lv.push_back( Vector_3::GetXAxis( ) + 2.0*Vector_3::GetYAxis( ) );
        lv.push_back( Vector_3::GetXAxis( ) + 3.0*Vector_3::GetYAxis( ) );
        const Line_3 lOut = BestLine( lv );
        if( (lOut.LineAxis( ) - Vector_3::GetYAxis( )).Norm( ) > 1.0E-8 )
        {
            ++g_errorCount;
            fprintf( stdout, "testLinePlane: BestLine-2a Failed \n" /*<< lOut.LineAxis( )*/ );
        }

        if( (lOut.BasePoint( )-Vector_3::GetXAxis( )).Norm( ) > 1.0E-8 )
        {
            ++g_errorCount;
            fprintf( stdout, "testLinePlane: BestLine-b Failed \n" /*<< lOut.BasePoint( )*/ );
        }
    }

    { /*test Distance from plane*/
        const Plane_3 p( Vector_3::GetXAxis( ), Vector_3(100.0, 0.0, 0.0) ); /*plane perpendicular to x-axis at 100 units along x*/
        const Vector_3 v( 90.0, 80.0, 10.0 );
        const double d1 = v.DistanceFromPlane( p );
        if( ::fabs( d1 - 10.0 ) > 1.0E-8 )
        {
            ++g_errorCount;
            fprintf( stdout, "testLinePlane: DistanceFromPlane failed\n" );
        }
    }

    { /*test Distance from line*/
        const Line_3 p( Vector_3::GetYAxis( ), Vector_3(100.0, 0.0, 0.0) ); /*line parallel to y-axis at 100 units along x*/
        const Vector_3 v( 90.0, 80.0, 2.0 );
        const double d1 = v.DistanceFromLine( p );
        if( ::fabs( d1 - ::sqrt( 104.0 ) ) > 1.0E-8 )
        {
            ++g_errorCount;
            fprintf( stdout, "testLinePlane: DistanceFromLine failed\n" );
        }
    }


}

/*=======================================================================*/
void testEigen( void )
{
    {
        double eigen1;
        Vector_3 evector1;

        std::vector<double> eigenValues;
        std::vector<Vector_3> eigenVectors;

        Matrix_3x3 m( 0.0,0.0,0.0, 0.0,0.0,0.0, 0.0,0.0,0.0 );
        m.Eigen1( eigen1, evector1 );
        m.Eigen3( eigenValues, eigenVectors );

        const std::vector<double> desiredValues( 3, 0.0 );

        if( eigen1 != 0.0 )
        {
            ++g_errorCount;
            fprintf( stdout, "testEigen: Eigen1 returned wrong eigenvalue for null matrix\n" );
        }

        if( eigenValues != desiredValues )
        {
            ++g_errorCount;
            fprintf( stdout, "testEigen: Eigen3 returned wrong eigenvalues eigenvalue for null matrix\n" );
        }
    }

    {
        double eigen1;
        Vector_3 evector1;

        std::vector<double> eigenValues;
        std::vector<Vector_3> eigenVectors;

        Matrix_3x3 m(UnitMatrix( ) );
        m.Eigen1( eigen1, evector1 );
        m.Eigen3( eigenValues, eigenVectors );

        const std::vector<double> desiredValues( 3, 1.0 );

        if( eigen1 != 1.0 )
        {
            ++g_errorCount;
            fprintf( stdout, "testEigen: Eigen1 returned wrong eigenvalue for unit matrix\n" );
        }

        if( eigenValues != desiredValues )
        {
            ++g_errorCount;
            fprintf( stdout, "testEigen: Eigen3 returned wrong eigenvalues for unit matrix\n" );
        }
    }

    {
        double eigen1;
        Vector_3 evector1;

        std::vector<double> eigenValues;
        std::vector<Vector_3> eigenVectors;

        Matrix_3x3 m( 3.,0.,0., 0.,-2.,0., 0.,0., 1. );
        m.Eigen1( eigen1, evector1 );
        m.Eigen3( eigenValues, eigenVectors );

        std::vector<double> desiredValues;
        desiredValues.push_back( 3.0 );
        desiredValues.push_back( -2.0 );
        desiredValues.push_back( 1.0 );

        if( eigen1 != 3.0 )
        {
            ++g_errorCount;
            fprintf( stdout, "testEigen: Eigen1 returned wrong eigenvalue for 3,-2,1 matrix\n" );
        }

        if( eigenValues != desiredValues )
        {
            ++g_errorCount;
            fprintf( stdout, "testEigen: Eigen3 returned wrong eigenvalues for 3,-2,1 matrix\n" );
        }

        if( eigenVectors[0] != xaxis )
        {
            ++g_errorCount;
            fprintf( stdout, "testEigen: Eigen3 returned wrong eigenvector[0] for 3,-2,1 matrix\n" );
        }

        if( eigenVectors[1] != yaxis )
        {
            ++g_errorCount;
            fprintf( stdout, "testEigen: Eigen3 returned wrong eigenvector[1] for 3,-2,1 matrix\n" );
        }

        if( eigenVectors[2] != zaxis )
        {
            ++g_errorCount;
            fprintf( stdout, "testEigen: Eigen3 returned wrong eigenvector[2] for 3,-2,1 matrix\n" );
        }
    }

    {
        double eigen1;
        Vector_3 evector1;

        std::vector<double> eigenValues;
        std::vector<Vector_3> eigenVectors;

        Matrix_3x3 m( 3.,0.,0., 0.,0.,0., 0.,0., 1. );
        m.Eigen1( eigen1, evector1 );
        m.Eigen3( eigenValues, eigenVectors );

        std::vector<double> desiredValues;
        desiredValues.push_back( 3.0 );
        desiredValues.push_back( 1.0 );
        desiredValues.push_back( 0.0 );

        if( eigen1 != 3.0 )
        {
            ++g_errorCount;
            fprintf( stdout, "testEigen: Eigen1 returned wrong eigenvalue for 3,0,1 matrix\n" );
        }

        if( eigenValues != desiredValues )
        {
            ++g_errorCount;
            fprintf( stdout, "testEigen: Eigen3 returned wrong eigenvalues for 3,0,1 matrix\n" );
        }

        if( eigenVectors[0] != xaxis )
        {
            ++g_errorCount;
            fprintf( stdout, "testEigen: Eigen3 returned wrong eigenvector[0] for 3,0,1 matrix\n" );
        }

        if( eigenVectors[1] != zaxis )
        {
            ++g_errorCount;
            fprintf( stdout, "testEigen: Eigen3 returned wrong eigenvector[1] for 3,0,1 matrix\n" );
        }

        if( eigenVectors[2] != -yaxis )
        {
            ++g_errorCount;
            fprintf( stdout, "testEigen: Eigen3 returned wrong eigenvector[2] for 3,0,1 matrix\n" );
        }
    }

    {
        double eigen1;
        Vector_3 evector1;

        std::vector<double> eigenValues;
        std::vector<Vector_3> eigenVectors;

        /* points are
            2,2,0
            -2,-2,0
            -1,1,0
            1,-1,0
            0,0,1
            0,0,-1

        sum x2+y2 = 20
        sum x2+z2 = 12
        sum y2+z2 = 12
        sum -xy = -6
        sum -xz = 0
        sum -yz = 0
        this is an ellipsoid of major diameter 4*root2 in the 1,1,0 direction
        minor diameter 2 in the 0,0,1 direction
        and semimajor 2*root2 in the 1,-1,0 direction (or reverse direction)*/
        Matrix_3x3 m(Matrix_3x3( 12., -6.,  0.,
                                 -6., 12.,  0.,
                                  0.,  0., 20. ) );

        m.Eigen1( eigen1, evector1 );
        m.Eigen3( eigenValues, eigenVectors );

        const double determinant2 = m.Det( );
        /*if the code is even close to correct, then the magnitudes of the next two
        terms must be almost the same. So dCompare does not need to use the
        larger of them, and Det is adequate.*/
        const double dTest = ::abs( determinant2 - eigenValues[0]*eigenValues[1]*eigenValues[2] );
        const double dCompare = ::abs( DBL_EPSILON*determinant2 );
        const double difference = dTest - dCompare;
        const double ratio = difference / determinant2;

        if( dTest > dCompare ) /* this could potentially fail with a small difference*/
        {
            ++g_errorCount;
            fprintf( stdout, "testEigen: Eigen3 failed for inertia tensor, difference=%g\n", difference );
        }

        if( ratio > 2.0*DBL_EPSILON ) /* this could potentially fail with a small difference*/
        {
            ++g_errorCount;
            fprintf( stdout, "testEigen: Eigen3 ratio=%g\n", ratio );
        }
    }
}

/*=======================================================================*/
void testUnitFunctions( void )
{
    Vector_3 vsmall( 0.01, 0.01, 0.02 );
    Vector_3 v1 = vsmall.UnitV( );
    Vector_3 v2 = UnitV( vsmall );

    if( (v1-v2).Norm( ) > 0.0 )
    {
        ++g_errorCount;
        fprintf( stdout, "testUnitFunctions: difference is wrong\n" );
    }

    if(::abs(v1[0]-v1[1]) > 0.0 || ::abs(v1[2]-2.0*v1[0]) > 0.0 )
    {
        ++g_errorCount;
        fprintf( stdout, "testUnitFunctions: member-A UnitV gave wrong value\n" );
    }

    if( ::abs(v2[0]-v2[1]) > 0.0 || ::abs(v2[2]-2.0*v2[0]) > 0.0 )
    {
        ++g_errorCount;
        fprintf( stdout, "testUnitFunctions: member-B UnitV gave wrong value\n" );
    }

    const Matrix_3x3 m = UnitMatrix( );
    const Matrix_3x3 mUnit( 1.0,0.0,0., 0.0,1.0,0.0, 0.0,0.0,1.0 );

    if( ::abs((m-mUnit).Det( )) > DBL_EPSILON )
    {
        ++g_errorCount;
        fprintf( stdout, "testUnitFunctions: UnitMatrix failed\n" );
    }

}

/*=======================================================================*/
void TestOnePair( const Vector_3& x1, const Vector_3& x2, const Vector_3& z1, const Vector_3& z2, int& localErrorCount )
{

    const double testValue1 = DBL_EPSILON * x1.Norm();
    const double testValue2 = DBL_EPSILON * x2.Norm();
    const double testValue3 = DBL_EPSILON * z1.Norm();
    const double testValue4 = DBL_EPSILON * z2.Norm();

    const double testMultiplier = 4.0;
    const double testLimit = testMultiplier*std::max( std::max( testValue1, testValue2 ), std::max( testValue3, testValue4 ) );
    const double testScalarLimit = testMultiplier*DBL_EPSILON;


    const Matrix_3x3 mPair = Vector_3::Pair( x1, x2, z1, z2 );
    const Vector_3 rotatedX1 = mPair.MV(x1);

    // test that x1 is in the z1,z2 plane
    const Vector_3 vtest1 = rotatedX1.UnitV( );
    const Vector_3 vtest2 = z1.UnitV( );
    const double vtest1Norm = vtest1.Norm( );
    const double vtest2Norm = vtest2.Norm( );

    const Vector_3 vDiff = vtest1 - vtest2;
    const double localTestLimit1 = 4.0*DBL_EPSILON*::sqrt(vtest1.Norm()*vtest2.Norm());

    if( vDiff.Norm( ) > localTestLimit1 ) // testing unit vectors
    { /* x1 should be rotated onto the z1 */
        ++localErrorCount;
        fprintf( stdout, "TestOnePair: x1!=z1 [1], vDiff.Norm() = %g\n", vDiff.Norm( ));
    }

    if( vtest1Norm-vtest2Norm > testScalarLimit ) // testing unit vectors
    { /* x1 should be rotated onto the z1 */
        ++localErrorCount;
        fprintf( stdout, "TestOnePair: Pair failed [1]\n");
    }
    // test that x2 is in the z1,z2 plane
    const double dDiff = z1.Cross(z2).UnitV().Dot(mPair.MV(x2));
    if( ::fabs(dDiff) > testLimit )
    {
        ++localErrorCount;
        fprintf( stdout, "TestOnePair: Pair failed [2]\n");
    }
    // test that the angle between x1 and z1 is essentially zero

    // other tests

    {
        const double distance1 = Plane_3( z1.Cross(z2), Vector_3::GetZeroVector( ) ) .DistanceFromPlane( rotatedX1 );
        if( distance1 > testLimit )
        {
            ++localErrorCount;
            fprintf( stdout, "TestOnePair: Pair distance1 not small enough, %g\n", distance1 );
        }
    }

    {
        const double angle1 = Vector_3::Angle( rotatedX1, Vector_3::GetZeroVector( ), z1 );
        if( ::fabs(angle1)  > testScalarLimit ) /*allow for some error*/
        {
            /*if this fails,x1 was not rotated onto zz1*/
            ++localErrorCount;
            fprintf( stdout, "TestOnePair: Pair failed for first rotation, angle1=%g radians\n", angle1 );
        }
    }

    const double angle2 = Vector_3::Torsion( x2.MV( mPair ), z1, Vector_3::GetZeroVector( ), z2 );
    const double distance2 = z1.Cross( z2 ).UnitV( ).Dot(x2.MV(mPair));
    if( distance2 > testLimit )
    {
        /*if this fails, x2 was not rotated into the plane of zz1,0,zz2*/
        ++localErrorCount;
        fprintf( stdout, "TestOnePair: Pair failed for second rotation distance, distance2=%g\n", distance2 );
    }
    else if( ::fabs(angle2)  > testScalarLimit ) /*allow for some error*/
    {
        ++localErrorCount;
        fprintf( stdout, "TestOnePair: Torsion failed for second rotation angle, angle2=%g radians\n", angle2 );
    }
}

void testCart( void )
{
   {   // Delaplane and Ibers, Acta Cryst B25, part 12, p 2423
      const double a=6.119;
      const double b=3.607;
      const double c=12.057;
      const double alpha = 90;
      const double beta  = 106.317;
      const double gamma = 90;

      const Vector_3 C1( -.0454, .0548, .0511 );
      const Vector_3 O1( .0848, -.0600, .1481 );

      const Matrix_3x3 m = Matrix_3x3::Cart( a,b,c, alpha,beta,gamma );

      const Vector_3 C1Cart( m.MV(C1) );
      const Vector_3 O1Cart( m.MV(O1) );

      const double C1ToO1Distance = (C1Cart-O1Cart).Norm( );

      if ( abs(C1ToO1Distance-1.285) > 0.002 )
      {
         ++g_errorCount;
         fprintf( stdout, "testCart: distance not good, %f\n", C1ToO1Distance );
      }
   }
   {   // Sabine, Fox, and Craven, Acta Cryst B25, part 12, p 2437
      const double a=6.119;
      const double b=3.607;
      const double c=12.057;
      const double alpha = 90;
      const double beta  = 106.317;
      const double gamma = 90;

      const Vector_3 C1( -.0452, .0547, .0510 );
      const Vector_3 O1( .0847, -.0599, .1488 );

      const Matrix_3x3 m = Matrix_3x3::Cart( a,b,c, alpha,beta,gamma );

      const Vector_3 C1Cart( m.MV(C1) );
      const Vector_3 O1Cart( m.MV(O1) );

      const double C1ToO1Distance = (C1Cart-O1Cart).Norm( );

      if ( abs(C1ToO1Distance-1.291) >0.002 )
      {
         ++g_errorCount;
         fprintf( stdout, "testCart: distance not good, %f\n", C1ToO1Distance );
      }
   }
   {
        // HEADER    OXIDOREDUCTASE                          26-FEB-09   3GFG
        // CRYST1   72.931   92.135  187.589  95.31  94.19  95.21 P 1          12
        // SCALE1      0.013712  0.001250  0.001136        0.00000
        // SCALE2      0.000000  0.010899  0.001093        0.00000
        // SCALE3      0.000000  0.000000  0.005372        0.00000
      const double a = 72.931;
      const double b = 92.135;
      const double c = 187.589;
      const double alpha = 95.31;
      const double beta  = 94.19;
      const double gamma = 95.21;
      const Matrix_3x3 inverseCartFromPDBFile( 
                                    0.013712,  0.001250,  0.001136,
                                    0.000000,  0.010899,  0.001093,
                                    0.000000,  0.000000,  0.005372
                                  );
      const Matrix_3x3 directCartFromLit = inverseCartFromPDBFile.Inver( );
      const Matrix_3x3 testCart = Matrix_3x3::Cart( a,b,c, alpha,beta,gamma );

      const Matrix_3x3 difference = testCart-directCartFromLit;
      const double detDifference = difference.Det( );
      if ( abs( detDifference ) > 0.001 )
      {
         ++g_errorCount;
         fprintf( stdout, "testCart: determinant difference too large\n" );
      }

      const double volumeRatio = directCartFromLit.Det( ) / testCart.Det( );

      // attempt to estimate the accuracy limit of the PDB inverse matrix (and some more for the cell)
      const double volumeRatioTest = ::fabs( volumeRatio - 1.0 ) / ( ::fabs(1.0/0.005373-1.0/0.005372)/0.005372 );
      if ( abs( volumeRatio - 1.0 ) > 10.0 * volumeRatioTest )
      {
         ++g_errorCount;
         fprintf( stdout, "testCart: volume ratio not 1.0, difference = %f\n", ::fabs(volumeRatio-1.0) );
         fprintf( stdout, "\n\n cart matrix differences\n %f %f %f\n %f %f %f\n %f %f %f\n\n",
            directCartFromLit[0] - testCart[0],
            directCartFromLit[1] - testCart[1],
            directCartFromLit[2] - testCart[2],
                                            
            directCartFromLit[3] - testCart[3],
            directCartFromLit[4] - testCart[4],
            directCartFromLit[5] - testCart[5],
                                            
            directCartFromLit[6] - testCart[6],
            directCartFromLit[7] - testCart[7],
            directCartFromLit[8] - testCart[8] 
         );


      }
   }
}


