#ifndef vector_3_INCLUDED
#define vector_3_INCLUDED

#ifdef __cplusplus
#endif

#include <algorithm>
#include <cfloat>
#include <cmath>
//#include "QRlib.h"
#include <utility>
#include <vector>

// forward declarations
class Plane_3;
class Line_3;
class Vector_3;
class Matrix_3x3;
template<typename DistanceType> class QR;

// these are friend functions, put here because LINT complains if they are at the end
//#ifdef __cplusplus
//    std::ostream& operator<<     ( std::ostream&, const Vector_3& );
//    std::ostream& operator<<     ( std::ostream&, const Matrix_3x3& );
//#endif
Vector_3 operator* ( const double&, const Vector_3& );
void initrn( int& iseed, int& indx, int& jndx, int& kndx, double buffer[] );
Matrix_3x3 RandomOrientation( int& iseed );
double randf( int& iseed );
double randfg( int& iseed );
Vector_3 randVector( int& iseed );
double Angle ( const Vector_3& a,  const Vector_3& b );


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
class Vector_3
{
    friend class Matrix_3x3;
    friend class QR<double>;

#ifdef __cplusplus
    friend std::ostream& operator<< ( std::ostream&, const Vector_3& );
#endif
    friend Vector_3 operator* ( const double&, const Vector_3& );
private:
    double v[3];

public:
    //CONSTRUCTORS, DESTRUCTOR
//This is a translation of an old fortran library of vector algebra code. It
//was designed to be versatile and easy to use. Surprisingly, even in 2005, it
//is in quite common use in the scientific community. 
//
//Conventions:
//   vector are internally arrays of 3 doubles
//   matrices are internally arrays of 9 doubles, indexed as follows
//      0 1 2
//      3 4 5
//      6 7 8
//

//-----------------------------------------------------------------------------
// Name: Vector_3()
// Description: default constructor provides bad-looking data
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
Vector_3( void )
//-------------------------------------------------------------------------------------
{
    v[0]=v[1]=v[2]=DBL_MAX;
}

//-----------------------------------------------------------------------------
// Name: Vector_3()
// Description: constructor to insert 3 values into a new Vector_3 vector
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
inline Vector_3( const double a, const double b, const double c )
//-------------------------------------------------------------------------------------
{
    v[0] = a;
    v[1] = b;
    v[2] = c;
}

//-----------------------------------------------------------------------------
// Name: Vector_3()
// Description: constructor to insert 3 identical values into a new Vector_3 vector
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
explicit Vector_3( const double a )
//-------------------------------------------------------------------------------------
{
    v[0] = a;
    v[1] = a;
    v[2] = a;
}

//-----------------------------------------------------------------------------
// Name: Vector_3()
// Description: copy constructor
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
Vector_3 ( const Vector_3& o )
//-------------------------------------------------------------------------------------
{
    v[0] = o.v[0];
    v[1] = o.v[1];
    v[2] = o.v[2];
}


//-----------------------------------------------------------------------------
// Name:~ Vector_3()
// Description: destructor
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
~Vector_3( void )
//-------------------------------------------------------------------------------------
{
    v[0]=v[1]=v[2]=DBL_MAX;
}
    //SCALAR OPERATIONS ON VECTORS
//-----------------------------------------------------------------------------
// Name: operator*()
// Description: multiply a vector by a floating point number
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline Vector_3 operator*  ( const double a ) const
//-----------------------------------------------------------------------------
    {
        return ( Vector_3(a*v[0],a*v[1],a*v[2]) );
    }

//-----------------------------------------------------------------------------
// Name: operator/()
// Description: divide a vector by a floating point number
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline Vector_3 operator/  ( const double a ) const
//-----------------------------------------------------------------------------
    {
        return ( Vector_3(v[0]/a,v[1]/a,v[2]/a) );
    }

//-----------------------------------------------------------------------------
// Name: operator+=()
// Description: add two vectors
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline Vector_3& operator+= ( const Vector_3& vv )
//-----------------------------------------------------------------------------
    {
        *this = *this + vv;
        return( *this );
    }

//-----------------------------------------------------------------------------
// Name: operator-=()
// Description: subtract two vectors
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline Vector_3& operator-= ( const Vector_3& vv )
//-----------------------------------------------------------------------------
    {
        (*this) = (*this) - vv;
        return( *this );
    }

    static double        Angle         ( const Vector_3& a,  const Vector_3& b,  const Vector_3& c );
    static double        Torsion        ( const Vector_3& a,  const Vector_3& b,  const Vector_3& c, const Vector_3& d );

    //OPERATIONS ONLY ON VECTORS
//-----------------------------------------------------------------------------
// Name: Dot()
// Description: compute the dot product of two vectors = |v1|*|v2|*cos(angle between)
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline double Dot ( const Vector_3& v_other ) const
//-----------------------------------------------------------------------------
    {
        return(
            v[0] * v_other[0]+
            v[1] * v_other[1]+
            v[2] * v_other[2]    );
    }

//-----------------------------------------------------------------------------
// Name: Cross()
// Description: compute the cross product of two vectors. The result is a vector
// perpendicular to the two and with Norm=|v1|*|v2|*sin(angle between)
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline Vector_3 Cross ( const Vector_3& v_other ) const
//-----------------------------------------------------------------------------
    {

        return (Vector_3(    v[1]*v_other[2]-v_other[1]*v[2],
                            -v[0]*v_other[2]+v_other[0]*v[2],
                             v[0]*v_other[1]-v_other[0]*v[1]  ) );
    }

//-----------------------------------------------------------------------------
// Name: DyadProduct()
// Description: compute the dyad product of two vectors
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    Vector_3 DyadProduct ( const Vector_3& v_other ) const
//-----------------------------------------------------------------------------
    {
        return ( Vector_3(v[0]*v_other[0],v[1]*v_other[1],v[2]*v_other[2]) );
    }
	
    Matrix_3x3 MatrixProduct ( const Vector_3& v_other ) const;

//-----------------------------------------------------------------------------
// Name: operator+()
// Description: add two vectors
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline Vector_3 operator+ ( const Vector_3& v_other ) const
//-----------------------------------------------------------------------------
    {
        return ( Vector_3( v[0]+v_other[0],v[1]+v_other[1],v[2]+v_other[2] ) );
    }

//-----------------------------------------------------------------------------
// Name: operator-()
// Description: subtract two vectors
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline Vector_3 operator- ( const Vector_3& v_other ) const
//-----------------------------------------------------------------------------
    {
        return ( Vector_3(v[0]-v_other[0],v[1]-v_other[1],v[2]-v_other[2]) );
    }

    //-----------------------------------------------------------------------------
    // Name: operator-()
    // Description: Unary minus for a vector
    //
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    Vector_3 operator- ( void ) const
    //-------------------------------------------------------------------------------------
    {
        return( Vector_3( -v[0], -v[1], -v[2] ) );
    }

    inline double Norm ( void ) const
    {
        return( sqrt( v[0]*v[0] + v[1]*v[1] + v[2]*v[2] ) );
    }


//-----------------------------------------------------------------------------
// Name: UnitV()
// Description: returns a vector that is the same as the input vector, but scaled to
// a unit vector
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
Vector_3 UnitV( void ) const
//-------------------------------------------------------------------------------------
{
    //fabs returns a double - test against something
    //considerably larger than DBL_MIN
    if ( ::fabs(v[0])+::fabs(v[1])+::fabs(v[2]) <10.0*FLT_MIN )
    {
        return (Vector_3::GetXAxis( ) );
    }
    else
    {
        return (Vector_3(*this)/Norm( ));  // const, so work on a copy
    }
}

    inline Vector_3& operator=  ( const Vector_3& v_other )
    {
        if (this != &v_other)
        {
            v[0] = v_other.v[0];
            v[1] = v_other.v[1];
            v[2] = v_other.v[2];
        }
        return (*this);
    }

    inline double SquaredLength ( void ) const
    {
        return( v[0]*v[0]+v[1]*v[1]+v[2]*v[2] );
    }

    static Matrix_3x3   Pair            ( const Vector_3& x1, const Vector_3& x2, const Vector_3& y1, const Vector_3& y2 );
    inline static Vector_3     GetXAxis ( void ) { return( Vector_3( 1.0, 0.0, 0.0 ) ); };
    inline static Vector_3     GetYAxis ( void ) { return( Vector_3( 0.0, 1.0, 0.0 ) ); };
    inline static Vector_3     GetZAxis ( void ) { return( Vector_3( 0.0, 0.0, 1.0 ) ); };
    inline static Vector_3     GetZeroVector ( void ) { return( Vector_3( 0.0, 0.0, 0.0 ) ); };

    bool operator== ( const Vector_3& v_other ) const { return( v[0]==v_other.v[0] && v[1]==v_other.v[1] && v[2]==v_other.v[2] ); };
    bool operator!= ( const Vector_3& v_other ) const { return( v[0]!=v_other.v[0] || v[1]!=v_other.v[1] || v[2]!=v_other.v[2] ); };
    static QR<double> QuaternionFromPair( const Vector_3& v1, const Vector_3& v2 );


//-----------------------------------------------------------------------------
// Name: GetCenterOfMass()
// Description: for an input list of points, compute the center of mass
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
template< typename ContainerType > 
static Vector_3 GetCenterOfMass( const ContainerType& lv )
//-------------------------------------------------------------------------------------
{
    Vector_3 cm(0.0);
    if( lv.empty( ) )
    {
        return( cm );
    }
    typename ContainerType::const_iterator it;
    for( it=lv.begin( ); it!=lv.end( ); ++it )
    {
        cm += (*it);
    }
    return( cm / static_cast<double>(lv.size( ) ) );
}
    //OPERATIONS USING MATRICES
    Vector_3 MV( const Matrix_3x3& m) const;

    //complex operations
    Matrix_3x3 Rotmat ( const double angle ) const;
    Matrix_3x3 InertiaTensorForPoint ( const double weight ) const;
    double DistanceFromPlane ( const Plane_3& p ) const;
    double DistanceFromLine ( const Line_3& v ) const;

    //access and io functions
    //-----------------------------------------------------------------------------
    // Name: operator[]()
    // Description: access function for the components of a vector
    //
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline double operator[] ( const int& i ) const
    //-------------------------------------------------------------------------------------
    {
        int n = ( i<0 ) ? 0 : i;
        if( i > 2 ) n = 2;
        return (v[n]);
    }

    inline double& operator[] ( const int& i )
    //-------------------------------------------------------------------------------------
    {
        int n = ( i<0 ) ? 0 : i;
        if( i > 2 ) n = 2;
        return (v[n]);
    }

    //-----------------------------------------------------------------------------
    // Name: at()
    // Description: access function for the components of a vector
    //
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline double at ( const int& i ) const
    //-------------------------------------------------------------------------------------
    {
        int n = ( i<0 ) ? 0 : i;
        if( i > 2 ) n = 2;
        return (v[n]);
    }

    // constants
    static const double    MINNORM;
}; // end of class vector

/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
class Matrix_3x3
{
    friend class Vector_3;
    friend class QR<double>;

#ifdef __cplusplus
    friend std::ostream& operator<< ( std::ostream&, const Matrix_3x3& );
#endif
    friend Matrix_3x3 operator* ( const double&, const Matrix_3x3& );
    double m[9];
public:
    //CONSTRUCTORS, DESTRUCTOR

//-----------------------------------------------------------------------------
// Name: Matrix_3x3()
// Description: constructor
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
Matrix_3x3( void )
//-------------------------------------------------------------------------------------
{
    for ( int i=0; i<9; i++) m[i] = DBL_MAX;
}

//-----------------------------------------------------------------------------
// Name: Matrix_3x3()
// Description: constructor to fill a matrix with a list of values
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
Matrix_3x3( const double a, const double b, const double c,
                        const double d, const double e, const double f,
                       const double g, const double h, const double i )
//-------------------------------------------------------------------------------------
{
    m[0] = a;
    m[1] = b;
    m[2] = c;
    m[3] = d;
    m[4] = e;
    m[5] = f;
    m[6] = g;
    m[7] = h;
    m[8] = i;
}

//-----------------------------------------------------------------------------
// Name: Matrix_3x3()
// Description: copy constructor
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
Matrix_3x3( const Matrix_3x3& o )
//-------------------------------------------------------------------------------------
{
    for ( int i=0; i<9; ++i) m[i] = o.m[i];
}

//-----------------------------------------------------------------------------
// Name: ~Matrix_3x3()
// Description: destructor
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
~Matrix_3x3( void )
//-------------------------------------------------------------------------------------
{
    for ( int i=0;i<9;i++) m[i] = DBL_MAX;
}
    //matrix-vector operations
//-----------------------------------------------------------------------------
// Name: MV()
// Description: multiply a vector (from the right) by a matrix
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
inline Vector_3 MV ( const Vector_3& v ) const
//-------------------------------------------------------------------------------------
{
    return (Vector_3( m[0]*v[0]+m[1]*v[1]+m[2]*v[2],
                      m[3]*v[0]+m[4]*v[1]+m[5]*v[2],
                      m[6]*v[0]+m[7]*v[1]+m[8]*v[2]));
}
    
    Vector_3 operator* ( const Vector_3& v ) const
    {
        return (Vector_3(m[0]*v[0]+m[1]*v[1]+m[2]*v[2],
        m[3]*v[0]+m[4]*v[1]+m[5]*v[2],
        m[6]*v[0]+m[7]*v[1]+m[8]*v[2]));
    }

    //operations on matrices only
    Matrix_3x3 operator*  ( const Matrix_3x3& mm ) const;

//-----------------------------------------------------------------------------
// Name: operator+()
// Description: add two matrices
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
inline Matrix_3x3 operator+ ( const Matrix_3x3& o ) const
//-------------------------------------------------------------------------------------
{
    return (Matrix_3x3(
        m[0]+o.m[0],
        m[1]+o.m[1],
        m[2]+o.m[2],
        m[3]+o.m[3],
        m[4]+o.m[4],
        m[5]+o.m[5],
        m[6]+o.m[6],
        m[7]+o.m[7],
        m[8]+o.m[8]  ));
}


//-----------------------------------------------------------------------------
// Name: operator-()
// Description: subtract two matrices
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
inline Matrix_3x3 operator- ( const Matrix_3x3& o ) const
//-------------------------------------------------------------------------------------
{
    return (Matrix_3x3(m[0]-o.m[0],m[1]-o.m[1],m[2]-o.m[2],
        m[3]-o.m[3],m[4]-o.m[4],m[5]-o.m[5],
        m[6]-o.m[6],m[7]-o.m[7],m[8]-o.m[8]));
}

//-----------------------------------------------------------------------------
// Name: operator*()
// Description: multiplies a matrix by a floating point number
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
inline Matrix_3x3 operator* ( const double d ) const
//-------------------------------------------------------------------------------------
{
    Matrix_3x3 mTemp(*this);
    for( int i=0; i<9; ++i )
    {
        mTemp.m[i] *= d;
    }
    return( mTemp );
}

//-----------------------------------------------------------------------------
// Name: operator/()
// Description: divides a matrix by a floating point number
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
Matrix_3x3 operator/ ( const double d ) const
//-------------------------------------------------------------------------------------
{
    Matrix_3x3 mTemp(*this);
    for( int i=0; i<9; ++i )
    {
        mTemp.m[i] /= d;
    }
    return( mTemp );
}

//-----------------------------------------------------------------------------
// Name: operator+=()
// Description: add two matrices
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
Matrix_3x3& operator+= ( const Matrix_3x3& mm )
//-------------------------------------------------------------------------------------
{
    (*this) = (*this) + mm;
    return( *this );
}

//-----------------------------------------------------------------------------
// Name: operator-=()
// Description: subtracts two matrices
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
Matrix_3x3& operator-= ( const Matrix_3x3& mm )
//-------------------------------------------------------------------------------------
{
    (*this) = (*this) - mm;
    return( *this );
}

//-----------------------------------------------------------------------------
// Name: operator*=()
// Description: multiplies two matrices
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
Matrix_3x3& operator*= ( const Matrix_3x3& mm )
//-------------------------------------------------------------------------------------
{
    (*this) = (*this) * mm;
    return( *this );
}

//-----------------------------------------------------------------------------
// Name: operator=()
// Description: copy a matrix
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
inline Matrix_3x3& operator= ( const Matrix_3x3& o )
//-------------------------------------------------------------------------------------
{
    if (this!=&o)
    {
        int i;
        for (i=0;i<9;i++) m[i] = o.m[i];
    }
    return (*this);
}


//-----------------------------------------------------------------------------
// Name: Transpose()
// Description: returns the transpose of the input matrix
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
Matrix_3x3 Transpose ( void ) const
//-------------------------------------------------------------------------------------
{
    return (Matrix_3x3(m[0],m[3],m[6],
        m[1],m[4],m[7],
        m[2],m[5],m[8]));
}


//-----------------------------------------------------------------------------
// Name: Inver()
// Description: returns the inverse of a matrix
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
Matrix_3x3 Inver ( void ) const
//-------------------------------------------------------------------------------------
{
    double d(this->Det( ));
    if( d == 0.0 ) d = 1.0;
    if( ::fabs( d ) < 1.0E-20 )
    {
        d = ( d < 0.0 ) ? -d : d;
    }
    const Vector_3 v0( this->m[0], this->m[1], this->m[2] );
    const Vector_3 v1( this->m[3], this->m[4], this->m[5] );
    const Vector_3 v2( this->m[6], this->m[7], this->m[8] );

    const Vector_3 vOut0 = v1.Cross( v2 );
    const Vector_3 vOut1 = v2.Cross( v0 );
    const Vector_3 vOut2 = v0.Cross( v1 );

    const Matrix_3x3 mTemp(
        vOut0.v[0], vOut1.v[0], vOut2.v[0],
        vOut0.v[1], vOut1.v[1], vOut2.v[1],
        vOut0.v[2], vOut1.v[2], vOut2.v[2] );  // computes the ? of a matrix
    return( mTemp/ d );
}


    static Matrix_3x3 Cart       ( const double a, const double b, const double c, 
                            const double alpha, const double beta, const double gamma );

    //-----------------------------------------------------------------------------
// Name: UnitMatrix()
// Description: return a unit matrix
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void UnitMatrix ( void )
//-------------------------------------------------------------------------------------
{
    *this =  Matrix_3x3(
        1.0,0.0,0.0,
        0.0,1.0,0.0,
        0.0,0.0,1.0    );
}

//-----------------------------------------------------------------------------
// Name: Det()
// Description: returns the determinant of a matrix
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
double Det( void ) const
//-------------------------------------------------------------------------------------
{
    const double d = m[0]*m[4]*m[8] + m[1]*m[5]*m[6] + m[2]*m[3]*m[7]
    -m[6]*m[4]*m[2] - m[7]*m[5]*m[0] - m[8]*m[3]*m[1];
    return( d );
}

//-----------------------------------------------------------------------------
// Name: Zero()
// Description: insert all zero values into a 3x3 matrix
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void Zero( void )
//-------------------------------------------------------------------------------------
{
    for( int i=0; i<9; ++i )
    {
        this->m[i] = 0.0;
    }
}

    bool Eigen1 ( double& eigenvalue, Vector_3& eigenvector1 ) const;
    bool Eigen3 ( std::vector<double>& eigenvalues, std::vector<Vector_3>& eigenvectors ) const;

//-----------------------------------------------------------------------------
// Name: operator[]()
// Description: access function for the components of a matrix
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
double operator[]( const int& i ) const
//-------------------------------------------------------------------------------------
{
    int n = ( i<0 ) ? 0 : i;
    if( i > 8 ) n = 8;
    return (m[n]);
}

double& operator[]( const int& i )
//-------------------------------------------------------------------------------------
{
    int n = ( i<0 ) ? 0 : i;
    if( i > 8 ) n = 8;
    return (m[n]);
}

    Matrix_3x3 UNX( void ) const;
    Matrix_3x3 UNY( void ) const;
    Matrix_3x3 UNZ( void ) const;

}; // end of class Matrix_3x3


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
// Line_3
//     This is a small class to represent a line. Minimal functionality in Vector_3 is
//     provided for working with lines and planes.
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
class Line_3
{
public:
    Line_3( void );
    Line_3( const Vector_3& direction, const Vector_3& CenterOfMass );
    Vector_3 LineAxis( void ) const;
    Vector_3 BasePoint( void ) const;
    virtual ~Line_3( void );
    double DistanceFromLine( const Vector_3& v ) const;

private:
    Vector_3 m_lineAxis;     // the vector parallel to the line
    Vector_3 m_basePoint;    // the point on the line where a vector from the origin intersects at 90 degrees
    Vector_3 m_centerOfMass; // center of mass of the points entered
}; // end of class Line_3

/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
// Plane_3
//     This is a small class to represent a plane. Minimal functionality in Vector_3 is
//     provided for working with lines and planes.
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
class Plane_3
{
public:
    Plane_3( void );
    Plane_3( const Vector_3& direction, const Vector_3& CenterOfMass);
    Vector_3 NormalVector( void ) const;
    Vector_3 BasePoint( void ) const;
    virtual ~Plane_3( void );
    double DistanceFromPlane( const Vector_3& v ) const;
    Line_3 Intersect( const Plane_3& p ) const;

private:
    Vector_3 m_planeNormal;  // the vector normal to the plane
    Vector_3 m_basePoint;    // the point on the plane where a vector from the origin intersects at 90 degrees
    Vector_3 m_centerOfMass; // center of mass of the points entered
}; // end of class Plane_3
template< typename DistanceType=double >
class QR
{

    friend class Vector_3;
    friend class Matrix_3x3;

private:
    DistanceType w, x, y, z;

public:
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline QR( ) : // default constructor
        w( DistanceType( 0.0 ) ),
        x( DistanceType( 0.0 ) ),
        y( DistanceType( 0.0 ) ),
        z( DistanceType( 0.0 ) )
    {
    }

    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline QR( const QR& q ) // copy constructor
    {
        if ( this != &q )
        {
            w = q.w;
            x = q.x;
            y = q.y;
            z = q.z;
        }
    }

    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline QR( const DistanceType& wi,
        const DistanceType& xi, 
        const DistanceType& yi, 
        const DistanceType& zi 
        ) :
        w( wi ),
        x( xi ),
        y( yi ),
        z( zi )
    {
    }

    /* QRSetQuaternion -- create an existing quaternion = w +ix+jy+kz */
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline void Set ( const DistanceType& wi,
        const DistanceType& xi, 
        const DistanceType& yi, 
        const DistanceType& zi 
        ) 
    {
        w = wi;
        x = xi;
        y = yi;
        z = zi;
    }

    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline DistanceType GetW( void ) const
    {
        return( w );
    }


    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline DistanceType GetX( void ) const
    {
        return( x );
    }


    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline DistanceType GetY( void ) const
    {
        return( y );
    }


    inline DistanceType GetZ( void ) const
    {
        return( z );
    }

    /* Add -- add a quaternion (q1) to a quaternion (q2) */   
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline QR operator+ ( const QR& q ) const
    {
        QR temp;
        temp.w = w + q.w;
        temp.x = x + q.x;
        temp.y = y + q.y;
        temp.z = z + q.z;
        return( temp );
    }

    /* Subtract -- subtract a quaternion (q2) from a quaternion (q1)  */  
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline QR operator- ( const QR& q ) const
    {
        QR temp;
        temp.w = w - q.w;
        temp.x = x - q.x;
        temp.y = y - q.y;
        temp.z = z - q.z;
        return( temp );
    }

    /* Multiply -- multiply a quaternion (q1) by quaternion (q2)  */    
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline QR operator* ( const QR& q ) const // multiply two quaternions
    {
        QR temp;
        temp.w = -z*q.z - y*q.y - x*q.x + w*q.w;
        temp.x =  y*q.z - z*q.y + w*q.x + x*q.w;
        temp.y = -x*q.z + w*q.y + z*q.x + y*q.w;
        temp.z =  w*q.z + x*q.y - y*q.x + z*q.w;
        return( temp );
    }

    /* Divide -- Divide a quaternion (q1) by quaternion (q2)  */
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    QR operator/ ( const QR& q2 ) const
    {
        const DistanceType norm2sq = q2.w*q2.w + q2.x*q2.x + q2.y*q2.y + q2.z*q2.z;

        if ( norm2sq == 0.0 )
        {
            return( QR( DBL_MAX, DBL_MAX, DBL_MAX, DBL_MAX ) );
        }

        QR q;
        q.w =  z*q2.z + y*q2.y + x*q2.x + w*q2.w;
        q.x = -y*q2.z + z*q2.y - w*q2.x + x*q2.w;
        q.y =  x*q2.z - w*q2.y - z*q2.x + y*q2.w;
        q.z = -w*q2.z - x*q2.y + y*q2.x + z*q2.w;

        return( q / norm2sq );
    }

    /* ScalarMultiply -- multiply a quaternion (q) by scalar (s)  */
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline QR operator* ( const DistanceType& d ) const // multiply by a constant
    {
        QR temp;
        temp.w = w*d;
        temp.x = x*d;
        temp.y = y*d;
        temp.z = z*d;
        return( temp );
    }

    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline QR operator/ ( const DistanceType& d ) const // divide by a constant
    {
        QR temp;
        temp.w = w/d;
        temp.x = x/d;
        temp.y = y/d;
        temp.z = z/d;
        return( temp );
    }

    /* Conjugate -- Form the conjugate of a quaternion qconj */
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline QR Conjugate ( void ) const
    {
        QR conjugate;
        conjugate.w =  w;
        conjugate.x = -x;
        conjugate.y = -y;
        conjugate.z = -z;
        return( conjugate );
    }

    /* Normsq -- Form the normsquared of a quaternion */
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline DistanceType Normsq ( void ) const
    {
        return( w*w + x*x + y*y + z*z );
    }

    /* Inverse -- Form the inverse of a quaternion */
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline QR Inverse ( void) const
    {
        QR inversequaternion = (*this).Conjugate();
        const DistanceType normsq = (*this).Normsq( );

        if ( normsq > DistanceType( 0.0 ) )
        {
            inversequaternion   = inversequaternion * DistanceType( 1.0 ) / normsq;
        }
        else
        {
            inversequaternion = QR( DBL_MAX, DBL_MAX, DBL_MAX, DBL_MAX );
        }
        return( inversequaternion );
    }

    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline QR& operator= ( const QR& q ) 
    {
        if ( this != &q )
        {
            w = q.w;
            x = q.x;
            y = q.y;
            z = q.z;
        }
        return( *this );
    }

    /* Equal */    
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline bool operator== ( const QR& q ) const
    {
        return( w==q.w && x==q.x && y==q.y && z==q.z );
    }

    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline bool operator!= ( const QR& q ) const
    {
        return( (w!=q.w) || (x!=q.x) || (y!=q.y) | (z!=q.z) );
    }

    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline Vector_3 operator* ( const Vector_3& v )
    {
        return( QRRotateByQuaternion( v ) );
    }

    /* RotateByQuaternion -- Rotate a vector by a Quaternion, w = qvq* */
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline Vector_3 QRRotateByQuaternion( const Vector_3& v )
    {
        QR vquat( 0.0, v[0], v[1], v[2] );
        const QR wquat = (*this)*vquat;
        const QR qconj = *this.Conjugate( );
        vquat = wquat * qconj;
        return( Vector_3( vquat.x, vquat.y, vquat.z ) );
    }

    /* Axis2Quaternion -- Form the quaternion for a rotation around axis v  by angle theta */
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    static inline QR QRAxis2Quaternion ( const Vector_3& v, const DistanceType& angle  )
    {
        const DistanceType norm2sq = v.SquaredLength();

        if ( norm2sq == DistanceType(0.0) )
        {
            return( QR( DBL_MAX, DBL_MAX, DBL_MAX, DBL_MAX ) );
        }

        const DistanceType sinOverNorm = sin(angle/2.0)/sqrt(norm2sq);
        const QR q( cos(angle/2.0), sinOverNorm*v[0], sinOverNorm*v[1], sinOverNorm*v[2]);

        return( q );
    }

    /* Matrix2Quaterion -- Form the quaternion from a 3x3 rotation matrix R */
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    static inline QR QRMatrix2Quaternion ( const Matrix_3x3& m )
    {
        const DistanceType trace = m[0] + m[4] + m[8];
        QR rotquaternion;

        if ( trace > -0.75 )
        {
            rotquaternion.w = sqrt( (1.0+trace)) / 2.0;
            const DistanceType recip = 0.25 / rotquaternion.w;
            rotquaternion.z = (m[3]-m[1]) * recip;
            rotquaternion.y = (m[2]-m[6]) * recip;
            rotquaternion.x = (m[7]-m[5]) * recip;
            return( rotquaternion );
        }

        const DistanceType fourxsq = 1.0 + m[0] - m[4] - m[8];

        if ( fourxsq >= 0.25 )
        {
            rotquaternion.x = sqrt( fourxsq ) / 2.0;
            const DistanceType recip = 0.25 /rotquaternion.x;
            rotquaternion.y = m[3] + m[1]*recip;
            rotquaternion.z = m[2] + m[6]*recip;
            rotquaternion.w = m[7] + m[5]*recip;
            return( rotquaternion );
        }

        const DistanceType fourysq = 1.0 + m[4] - m[0] - m[8];

        if ( fourysq >= 0.25 )
        {
            rotquaternion.y = sqrt( fourysq ) / 2.0;
            const DistanceType recip = 0.25 / rotquaternion.y;
            rotquaternion.x = m[3] + m[1]*recip;
            rotquaternion.w = m[2] + m[6]*recip;
            rotquaternion.z = m[7] + m[5]*recip;
            return( rotquaternion );
        }

        const DistanceType fourzsq = 1.0 + m[8] - m[0] - m[4];

        if ( fourzsq >= 0.25 )
        {
            rotquaternion.z = sqrt( fourzsq ) / 2.0;
            const DistanceType recip = 0.25 / rotquaternion.z;
            rotquaternion.w = m[3] - m[1]*recip;
            rotquaternion.w = m[2] - m[6]*recip;
            rotquaternion.w = m[7] - m[6]*recip;
            return( rotquaternion );
        }

        return( QR( ) );
    }

    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    DistanceType operator[] ( const int k ) const
    {
        const int i = (k<0) ? 0 : ( (k>3)?3:k ) ;
        if ( i==0 ) return w;
        if ( i==1 ) return x;
        if ( i==2 ) return y;
        if ( i==3 ) return z;
        return( 0 ); // just to keep compilers happy
    }

    /* Quaternion2Matrix -- Form the 3x3 rotation matrix from a quaternion */    
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline Matrix_3x3 QRQuaternion2Matrix( void ) const
    {
        const DistanceType ww = w*w;
        const DistanceType xx = x*x;
        const DistanceType yy = y*y;
        const DistanceType zz = z*z;

        const DistanceType  twoxy = 2.0 * x*y;
        const DistanceType  twoyz = 2.0 * y*z;
        const DistanceType  twoxz = 2.0 * x*z;
        const DistanceType  twowx = 2.0 * w*x;
        const DistanceType  twowy = 2.0 * w*y;
        const DistanceType  twowz = 2.0 * w*z;

        const Matrix_3x3 m(
            ww+xx-yy-zz,
            twoxy - twowz,
            twoxz + twowy,

            twoxy + twowz,
            ww-xx+yy-zz,
            twoyz - twowx,

            twoxz - twowy,
            twoyz + twowx,
            ww-xx-yy+zz
            );

        return( m );
    } // end QRQuaternion2Matrix

    /* Get a unit quaternion from a general one */
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline QR UnitQ( void ) const
    {
        const DistanceType normsq = Normsq( );
        if ( normsq == 0.0 )
        {
            return( QR( DBL_MAX, DBL_MAX, DBL_MAX, DBL_MAX ) );
        }
        else
        {
            return( (*this) / sqrt( Normsq( ) ) );
        }
    } // end UnitQ

    /* Quaternion2Angles -- Convert a Quaternion into Euler Angles for Rz(Ry(Rx))) convention */  
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    inline bool QRQuaternion2Angles ( DistanceType& rotX, DistanceType& rotY, DistanceType& rotZ ) const
    {
        const DistanceType PI = 4.0 * atan( 1.0 );

        const DistanceType rmx0 = w*w + x*x - y*y - z*z;
        const DistanceType rmx1 = 2.0 * ( x*y - w*z );
        const DistanceType rmy0 = 2.0 * ( x*y + w*z );
        const DistanceType rmy1 = w*w - x*x + y*y - z*z;
        const DistanceType rmz0 = 2.0 * (x*z - w*y );
        const DistanceType rmz1 = 2.0 * (w*x + y*z );
        const DistanceType rmz2 = w*w - x*x - y*y + z*z;
        DistanceType srx, sry, srz, trX, trY, trZ;

        if ( rmz0 >= 1.0 )
        {
            sry = -0.5 * PI;
        }
        else if ( rmz0 <= -1.0 )
        {
            sry = 0.5 * PI;
        }
        else
        {
            sry = asin( -rmz0 );
        }

        if ( rmz0 > 0.9999995 )
        {
            srx = atan2( -rmx1, rmy1 );
            srz = 0.0;
        }
        else if ( rmz0 < -0.9999995 )
        {
            srx = atan2( rmx1, rmy1 );
            srz = 0.0;
        }
        else
        {
            srx = atan2( rmz1, rmz2 );
            srz = atan2( rmy0, rmx0 );
        }

        trX = PI + srx;
        if ( trX > 2.0 * PI ) trX -= 2.0 * PI;
        trY = PI + sry;
        if ( trY > 2.0 * PI ) trY -= 2.0 * PI;
        trZ = PI + srz;
        if ( trZ > 2.0 * PI ) trZ -= 2.0 * PI;

        const DistanceType nsum  =    fabs( cos(srx)-cos(rotX)) + fabs( sin(srx)-sin(rotX))
                                    + fabs( cos(sry)-cos(rotY)) + fabs( sin(sry)-sin(rotY))
                                    + fabs( cos(srz)-cos(rotZ)) + fabs( sin(srz)-sin(rotZ));

        const DistanceType tsum  =    fabs( cos(trX)-cos(rotX)) + fabs( sin(trX)-sin(rotX))
                                    + fabs( cos(trY)-cos(rotY)) + fabs( sin(trY)-sin(rotY))
                                    + fabs( cos(trZ)-cos(rotZ)) + fabs( sin(trZ)-sin(rotZ));

        if ( nsum < tsum )
        {
            rotX = srx; rotY = sry; rotZ = srz;
        }
        else
        {
            rotX = trX; rotY = trY; rotZ = trZ;
        }

        return( true );
    }  // end QRQuaternion2Angles

    /* Angles2Quaternion -- Convert Euler Angles for Rz(Ry(Rx))) convention into a quaternion */
    /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    static inline QR QRAngles2Quaternion ( const DistanceType& rotX, const DistanceType& rotY, const DistanceType& rotZ )
    {
        const DistanceType cx = cos( rotX / 2.0 );
        const DistanceType sx = sin( rotX / 2.0 );

        const DistanceType cy = cos( rotY / 2.0 );
        const DistanceType sy = sin( rotY / 2.0 );

        const DistanceType cz = cos( rotZ / 2.0 );
        const DistanceType sz = sin( rotZ / 2.0 );

        const QR q( cx*cy*cz + sx*sy*sz,
            sx*cy*cz - cx*sy*sz,
            cx*sy*cz + sx*cy*sz,
            cx*cy*sz - sx*sy*cz 
            ); 
        return( q );
    } // end QRAngles2Quaternion
}; // end class QR


/*=========================================================================================*/
/* start of non-class functions */
/*=========================================================================================*/
static Vector_3 UnitV( const Vector_3& v ) { return( v/v.Norm( ) ); }
static Matrix_3x3 UnitMatrix( void ) { return( Matrix_3x3( 1.0,0.0,0.0, 0.0,1.0,0.0, 0.0,0.0,1.0 ) ); }

//-----------------------------------------------------------------------------
// Name: BestPlane()
// Description: for a list of points, the best plane is computed and returned.
//
//  The algorithm is that of Andrews and Bernstein, 1976, Acta Cryst, A32, 504-506
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
template< typename ContainerType > 
Plane_3 BestPlane ( const ContainerType& lv )
//-------------------------------------------------------------------------------------
{
    const Vector_3 CenterOfMass( Vector_3::GetCenterOfMass( lv ) );
    Matrix_3x3 inertiaTensor( 0.0,0.0,0.0, 0.0,0.0,0.0, 0.0,0.0,0.0 );
    typename ContainerType::const_iterator it;

    for( it=lv.begin( ); it!=lv.end( ); ++it )
    {
        inertiaTensor += ( (*it) - CenterOfMass ).InertiaTensorForPoint( 1.0 );
    }

    double eigenValue1;
    Vector_3 eigenVector1;
    inertiaTensor.Eigen1( eigenValue1, eigenVector1 );

    return( Plane_3( eigenVector1, CenterOfMass ) );
}

//-----------------------------------------------------------------------------
// Name: BestLine()
// Description: for a list of points, the best line is computed and returned.
//
//  The algorithm is that of Andrews and Bernstein, 1976, Acta Cryst, A32, 504-506
//
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
template< typename ContainerType > 
static Line_3 BestLine( const ContainerType& lv )
//-------------------------------------------------------------------------------------
{
    const Vector_3 CenterOfMass( Vector_3::GetCenterOfMass( lv ) );
    Matrix_3x3 inertiaTensor( 0.0,0.0,0.0, 0.0,0.0,0.0, 0.0,0.0,0.0 );
    typename ContainerType::const_iterator it;

    for( it=lv.begin( ); it!=lv.end( ); ++it )
    {
        inertiaTensor += ( (*it) - CenterOfMass ).InertiaTensorForPoint( 1.0 );
    }

    double eigenValue1;
    Vector_3 eigenVector1;
    inertiaTensor.Inver( ).Eigen1( eigenValue1, eigenVector1 );

    return( Line_3( eigenVector1, CenterOfMass ) );
}

/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
template< typename ContainerType >
static QR<double> AB_BestRotation( 
                              const ContainerType& v1, 
                              const ContainerType& v2,
                              Vector_3& centerOfMass1,
                              Vector_3& centerOfMass2,
                              Matrix_3x3& rotationMatrixFrom2To1,
                              double& rmsd,
                              const int pairCount=0 )
//-------------------------------------------------------------------------------------
{
   ContainerType centered_v1;
   ContainerType centeredRotated_v2;
   ContainerType axisBaseVectorList;
   ContainerType axisVectorList;
   std::vector<QR<double> > quaternionList;

      std::vector<long> serialList1( v1.size( ), 0 );
      std::vector<long> serialList2( v1.size( ), 0 );
      std::vector<std::pair< std::pair<long,long>, std::pair<long,long> > > serials( v1.size( ), std::make_pair( std::make_pair(0,0),std::make_pair(0,0) ) );
      std::vector<std::pair<long,long> > indices;
      const int singleAtomSelect = 0;

       QR<double> quaternion =  AB_BestRotationCore(
                     v1,
                     v2,
                     centerOfMass1,
                     centerOfMass2,
                     serialList1,  //5
                     serialList2,
                     centered_v1,
                     centeredRotated_v2,
                     serials,
                     indices, // 10
                     quaternionList,
                     axisVectorList,
                     axisBaseVectorList,
                     rotationMatrixFrom2To1,
                     rmsd,
                     pairCount,
                     singleAtomSelect
                     );
   return( quaternion );
}

/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    // INPUT
    // v1                       The Cartesian coordinates of molecule1.
    // v2                       The Cartesian coordinates of molecule2.
    //
    // OUTPUT
    // centerOfMass1            The center of mass of molecule1.
    // centerOfMass2            The center of mass of molecule2.
    // serial1                  The PDB atom serial numbers for molecule1.
    // serial2                  The PDB atom serial numbers for molecule2.
    // centered_v1              The coordinates of molecule1, translated so its center of mass is at the origin.
    // centeredRotated_v2       The coordinates of molecule2 in the position to best match molecule1 after
    //                          each has been moved to the origin (and molecule2 rotated).
    // serials                  The pair of atom numbers from the PDB file for the two atom pairs used
    //                          to calculate an axis.
    // indices                  The pair of indices used for a particular axis calculations. This is 
    //                          the indices in v1 and v2.
    // quaternionList           The quaternion for each pair that rotates the first onto the second
    // axisVectorList           The direction vector of each axis
    // axisBaseVectorList       The point on each axis vector nearest the origin
    // rotationMatrixFrom2To1   Matrix that will rotate molecule2 (as well as possible) onto molecule1
    // rmsd                     The root mean sq. deviation between the two fitted molecules
    // pairCount=0              The number of pairs to calculate for each atom in the first molecule. The second pair
    //                          will be chose randomly. This is here so that the number can be limited 
    //                          in the case of a large dataset. 
    //                          ALSO, The value zero means do the complete exhaustive list, but each one will only
    //                          be done once.
    //                          ALSO, A negative value means use the next parameter to select a particular
    //                          atom to be the first pair and do ALL of the other pairs against it.
    // singleAtomSelect         Chooses the single atom (and its rotated mate) and looks at all the other
    //                          pairs compared to that one. ONLY used if pairCount is negative!!!!!
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
template< typename ContainerTypeV, typename ContainerTypeQ >
static QR<double> AB_BestRotationCore( 
                              const ContainerTypeV& v1, 
                              const ContainerTypeV& v2,
                              Vector_3& centerOfMass1,
                              Vector_3& centerOfMass2,
                              std::vector<long>& serial1, // 5
                              std::vector<long>& serial2,
                              ContainerTypeV& centered_v1,
                              ContainerTypeV& centeredRotated_v2,
                              std::vector<std::pair< std::pair<long,long>, std::pair<long,long> > >& serials,
                              std::vector<std::pair<long,long> >& indices, // 10
                              ContainerTypeQ& quaternionList,
                              ContainerTypeV& axisVectorList,
                              ContainerTypeV& axisBaseVectorList,
                              Matrix_3x3& rotationMatrixFrom2To1,
                              double& rmsd,
                              const int pairCount=0,
                              const int singleAtomSelect=-INT_MAX
                              )
//-------------------------------------------------------------------------------------
{

   ContainerTypeV vlist1(v1);
   ContainerTypeV vlist2(v2);

   const int chainACount = vlist1.size( );
   const int chainBCount = vlist2.size( );

   QR<double> q( 0.0, 0.0, 0.0, 0.0 ); // Quaternion
   QR<double> qsum( 0.0, 0.0, 0.0, 0.0 ); // Quaternion sum for averaging

   Vector_3 axisSum(0.0,0.0,0.0);

   int nvec;
   if ( chainACount != chainBCount )
   {
      return( q );
   }
   else
   {
      nvec = chainACount;
   }

   centerOfMass1 = Vector_3::GetCenterOfMass( vlist1 ); // RETURN PARAMETER
   centerOfMass2 = Vector_3::GetCenterOfMass( vlist2 ); // RETURN PARAMETER

   // subtract the centers of mass
   for ( int i=0; i<nvec; ++i )
   {
      vlist1[i] -= centerOfMass1;
      vlist2[i] -= centerOfMass2;
   }

   std::vector<Plane_3> plist(nvec);
   for ( int i=0; i<nvec; ++i )
   {
      plist[i] = Plane_3( vlist2[i]-vlist1[i], 0.5*(vlist2[i]+vlist1[i]) );
   }

   long count = 0;
   // do each atom

   {
      int outerLoopStart;
      int outerLoopEnd;
      if ( pairCount >= 0 )
      {
         outerLoopStart = 0;
         outerLoopEnd   = nvec;
      }
      else
      {
         outerLoopStart = singleAtomSelect;
         outerLoopEnd   = singleAtomSelect+1;
      }

      for ( int i=outerLoopStart; i<outerLoopEnd; ++i )
      {
         int pairStart;
         int numberOfPairsPerAtom;
         if ( pairCount > 0 ) // choose random 2nd atom, override singleAtomSelection
         {
            pairStart = 0; // look at all the atoms in the target, but random selections
            numberOfPairsPerAtom = pairCount;
         }
         else if ( pairCount < 0 )
         {
            // look at all the atoms in the target, but from only one in the source 
            pairStart = 0;
            numberOfPairsPerAtom = nvec;
         }
         else // pairCount == 0  exhaustive case, but only one per pair (don't look back)
         {
            //  never look at the pair to itself
            pairStart = i+1;
            numberOfPairsPerAtom = nvec;
         }


         //const int pairStart = (pairCount==0) ? i+1 : 0; // pre singleAtomSelect code (for reference)
         //const int numberOfPairsPerAtom = (pairCount==0) ? nvec : pairCount;
         int index = 0;
         for ( int pairs=pairStart; pairs<numberOfPairsPerAtom; ++pairs )
         {
            if ( pairs == i )
            {
               continue;
            }

            // in the exhaustive case, are we done (last atom, last match are the same)
            if ( pairCount==0 && i>=nvec-1 && pairs>=nvec-1 )
            {
               break;
            }

            // pick the other pair of points
            if ( pairCount <= 0 )
            {
               index = pairs;
            }
            else
            {
               // find some random pair match that is not the same as this one
               while ( index=rand()*nvec/RAND_MAX, index>=nvec && pairs==index )
               {
               }
            }

            Line_3 l = plist[i].Intersect( plist[index] );

            if (l.LineAxis().Norm() == 0.0 )
            {
               continue;
            }

            if ( count>1 && l.LineAxis( ).Dot( axisSum ) < 0.0 )
            {
               l = Line_3( -l.LineAxis( ), l.BasePoint( ) );
            }

            double angle = Vector_3::Angle( 
                                                l.LineAxis().Cross( vlist1[i]), 
                                                Vector_3::GetZeroVector(), 
                                                l.LineAxis().Cross( vlist2[i])
                                                );


            q = QR<double>(  
                            cos(angle/2.0),
                            sin(angle/2.0) * l.LineAxis( )[0],
                            sin(angle/2.0) * l.LineAxis( )[1],
                            sin(angle/2.0) * l.LineAxis( )[2]
                          );
            if ( q[1]+q[2]+q[3] != 0 )
            {
               qsum = qsum + q;
               quaternionList.push_back( q );
               axisBaseVectorList.push_back( l.BasePoint( ) );
               axisSum += l.LineAxis( );
               axisVectorList.push_back( l.LineAxis( ) );
               serials.push_back( std::make_pair( std::make_pair( serial1[i], serial2[i] ), std::make_pair( serial1[index], serial2[index] ) ) );
               indices.push_back( std::make_pair( i, index ) );
           }


            const double basenorm = l.BasePoint().Norm();

            if( l.BasePoint().Norm() > 10000.0 )
            {
               const int i19191 = 19191;
            }

            //fprintf( stdout, " %d %d %f  %f  %f  %f  %f  %f  %f  %f\n", serial1[i], serial2[index],
            //   (vlist1[i]-vlist1[index]).Norm(), angle*180/4/atan(1.0), 
            //   l.LineAxis ()[0], l.LineAxis ()[1], l.LineAxis ()[2],
            //   l.BasePoint()[0], l.BasePoint()[1], l.BasePoint()[2]);
            ++count;
         }
      }
   }

   axisSum = axisSum.UnitV( ); // RETURN PARAMETER

   const double qnormsq = qsum[0]*qsum[0] + qsum[1]*qsum[1] + qsum[2]*qsum[2] + qsum[3]*qsum[3];

   if ( qnormsq == 0.0 || (qsum[1]+qsum[2]+qsum[3]) == 0.0 )
   {
      qsum = QR<double> ( 0.0, 0.0, 0.0, 0.0 );
      rmsd = DBL_MAX;
   }
   else
   {
      const double qnorm = sqrt( qnormsq );
      qsum = qsum / qnorm;
   }

   // Calculate the standard deviation of the atom fits
   double sum = 0.0;
   const double cosTerm = qsum.GetW( );
   const double sinTerm = sqrt( 1.0 - cosTerm*cosTerm );
   const double angleDegrees = -2.0*atan2( sinTerm, cosTerm ) * 180.0 / 4.0 / atan(1.0);
   //fprintf( stdout, "angleDegrees %f\n", angleDegrees );
   rotationMatrixFrom2To1 = axisSum.Rotmat( -2.0*atan2( sinTerm, cosTerm ) ); // RETURN PARAMETER

   centered_v1        = vlist1;
   centeredRotated_v2.resize( vlist2.size( ) );

   for( int i=0; i<nvec; ++i )
   {
      centeredRotated_v2[i] = vlist2[i].MV(rotationMatrixFrom2To1);
      sum += (vlist1[i]-centeredRotated_v2[i]) .SquaredLength( );
   }
   rmsd = sqrt(sum/nvec); // RETURN PARAMETER

   return( qsum );
}

std::pair<double,Vector_3> CircleCenterFromThreePoints( const Vector_3& v1, const Vector_3& v2, const Vector_3& v3 );


#endif  //  vector_3_INCLUDED

